import types 
import time
import warnings
from collections import namedtuple

def mock(*kargs, **kwargs):
    things = {}
    attributes = []
    if 'template' in kwargs.keys():
        template = kwargs['template']
        items = [ method for method in dir(template) if not method.startswith('__')]
        method_types = [types.MethodType, types.FunctionType, types.BuiltinMethodType]

        methods = filter(lambda thing:type(getattr(template, thing)) in method_types , items)
        attributes = list(set(items).difference(methods))
        things = { method_name:MethodMock() for method_name in methods }
        things.update({attribute_name:AttributeMock(attribute_name) for attribute_name in attributes if not attribute_name.startswith('__')})

    if 'methods' in kwargs.keys():
        additional_methods = kwargs['methods']
        if isinstance(additional_methods, types.ListType):
            additional_methods= { method_name: None for method_name in additional_methods }

        things.update({ method_name:MethodMock(return_value) for method_name, return_value in additional_methods.items() })

    if 'attributes' in kwargs.keys():
        additional_attributes= kwargs['attributes']
        if isinstance(additional_attributes, types.ListType):
            additional_attributes= { attr_name: None for attr_name in additional_attributes }
        things.update({attr_name:AttributeMock(attr_name, attr_return) for attr_name, attr_return in additional_attributes.items()})
        attributes = list(set(attributes + additional_attributes.keys()))

    things['attribute_names'] = attributes
    return type("MockObject", (MockObject,), things)

class MockObject(object):
    def __init__(self, methods=[], attributes=[]):

        self.attribute_calls = []
        #self.__dict__.update({attribute_name:AttributeMock(attribute_name, None)
        #    for attribute_name in attributes})

    def __getattr__(self,name):
        warnings.warn('creating mock for method "'+name+'" that doesnt exsist', RuntimeWarning)
        mock_created = MethodMock()
        setattr(self, name,  mock_created )
        return mock_created

    @classmethod
    def methods(self):
        return dir(self)

    @property
    def all_calls(self):
        all_calls = self.attribute_calls + self.method_calls
        all_calls = sorted(all_calls, key=lambda c: c.time)
        return all_calls

    @property
    def method_calls(self):
        print 'calling all method calls'

        ignore_these = ['attribute_calls', 'method_calls', 'all_calls'] + self.attribute_names
        method_names = [method_name for method_name in dir(self)
                if method_name not in ignore_these  ]

        mock_methods_names = filter(lambda method_name: isinstance(getattr(self, method_name), MethodMock) , method_names)

        all_mock_calls = []
        for mock_method_name in mock_methods_names:
            local_mock_method = getattr(self, mock_method_name)
            for call in local_mock_method.calls:
                call.name = mock_method_name
            all_mock_calls += local_mock_method.calls

        all_mock_calls = sorted(all_mock_calls,
                key=lambda a_call: a_call.time)

        return all_mock_calls

class AttributeMock(object):

    def __init__(self, name, default=None):
        self.name = name
        self.value = default

    def _call_attributes(self, new_value):
        return {
            'name': self.name,
            'type': 'attribute',
            'time': time.clock(),
            'initial_value': self.value,
            'final_value': new_value,
        }

    def __set__(self, instance, value):
        call_attributes = {
            'name': self.name,
            'type': 'attribute',
            'time': time.clock(),
            'initial_value': self.value,
            'final_value': value,
        }

        instance.attribute_calls.append(type('MockCall', (object,), call_attributes))
        self.value = value

    def __get__(self, instance, owner_class):
        #print 'getting attribute ' + self.name
        call_attributes = {
            'name': self.name,
            'type': 'attribute',
            'time': time.clock(),
            'initial_value': self.value,
            'final_value': self.value,
        }

        instance.attribute_calls.append(type('MockCall', (object,), call_attributes))
        return self.value

class MethodMock():

    def __init__(self, returns=None):
        self.returns = returns
        self.calls = []

    def __call__(self, *args, **kwargs):
        call_attributes = {
            'type': 'method',
            'time': time.clock(),
            'args':args,
            "kwargs": kwargs
        }

        self.calls.append(type('MockCall', (object,), call_attributes))
        return self.returns

    @classmethod
    def returns(self, a_value):
        self.returns = a_value

    @classmethod
    def was_called(self):
        return len(self.calls) > 0




#import mock_these.dumb_class
from tests.mock_these.dumb_class import DumbClass
import warnings
import types
#warnings.warn(str([ name for name, type in locals().items() if  isinstance(type, types.ModuleType) ]))
#warnings.warn(str(id(locals())))
class ParentClass():
    def __init__(self):
        self.dumb = DumbClass()

    def get_attribute(self):
        return dumb.a_attribute

    def get_method(self):
        return self.dumb.a_method()

    def get_property(self):
        return self.dumb.a_property

    def get_class_method(self):
        return self.dumb.class_method()

    def get_staticmethod(self):
        return self.dumb.static_method()





from unittest import TestCase
from datetime import datetime
import patch
from patch import PatchObject
from object import MockObject, MethodMock
from tests.mock_these.dumb_class import DumbClass
from tests.mock_these.parent_class import ParentClass
import tests.mock_these.bunch_of_functions as bof
import warnings
import types
#warnings.warn(str([ name for name, type in locals().items() if  isinstance(type, types.ModuleType) ]))
#warnings.warn(str(id(locals())))

class TestPatchFunction(TestCase):
    @patch.function("tests.mock_these.bunch_of_functions.a_function")
    def test_patch_should_replace_function_with_methoond_mock(self, function_mock):
        self.assertIsInstance(bof.a_function, MethodMock)

    @patch.function("tests.mock_these.bunch_of_functions.a_function")
    def test_patch_function_returns_non_by_default(self, function_mock):
        self.assertEqual(bof.a_function(), None)

    @patch.function("tests.mock_these.bunch_of_functions.a_function", returns="new_value")
    def test_patch_should_replace_function_with_return_value(self, function_mock):
        self.assertEqual(bof.a_function(), "new_value")

class TestPatchObject(TestCase):
    def test_patch_should_not_effect_method_before(self):
        from datetime import datetime
        self.assertIsInstance(datetime.now(), datetime)

    @patch.object('datetime.datetime')
    def test_patched_object_should_return_patch_object(self, patch):
        from datetime import datetime
        self.assertIsInstance(datetime, PatchObject)

    @patch.object('datetime.datetime')
    def test_calling_patched_object_returns_mock_object(self, patch):
        from datetime import datetime
        self.assertIsInstance(datetime(), MockObject)

    @patch.object('datetime.datetime')
    def test_calling_method_of_class_object_returns_method_mock(self, patch):
        from datetime import datetime
        self.assertIsInstance(datetime.now, MethodMock)

    @patch.object('datetime.datetime')
    def test_patched_method_should_have_default_methods_of_mocked_class(self, patch):
        from datetime import datetime
        print dir(datetime)
        self.assertTrue('now' in  dir(datetime))

    @patch.object('datetime.datetime')
    def test_treat_all_methods_found_as_classmethods(self, patch):
        from datetime import datetime
        datetime.now()

    @patch.object('datetime.datetime')
    def test_adds_instances_to_patch_attribute(self, patch):
        from datetime import datetime
        mock_object = datetime()
        self.assertEqual(len(patch.instances), 1)
        self.assertEqual(mock_object, patch.instances[0])

    def test_patch_should_not_effect_method_after(self):
        from datetime import datetime
        self.assertIsInstance(datetime.now(), datetime)

    @patch.object('tests.mock_these.parent_class.DumbClass', methods={"a_method":"test return"})
    def test_patch_with_parent_has_a_relationship(self, dumb_patch):
        pc = ParentClass()
        #print globals()
        #print pc.dumb.a_method
        self.assertEqual(pc.get_method(), "test return")




from object import  mock, MethodMock
import warnings
import importlib

class PatchObject(object):
#, original_object,  attributes, methods
    def __init__(self, *args, **kwargs):
        self.mock_class = mock(*args, **kwargs)
        self.class_methods = {}
        self.instances = []

    def __dir__(self):
        return dir(self.mock_class)

    def __call__(self, *kargs, **kwargs):
        new_mock = self.mock_class()
        self.instances.append(new_mock)
        return new_mock

    def __getattr__(self, name):
        if name.startswith("__"):
            return getattr(self.mock_class, name)

        if name not in self.class_methods.keys():
            self.class_methods.update({name: MethodMock()})

        return self.class_methods[name]

    @property
    def class_method_calls(self):
        all_class_method_calls = []
        for class_method_name, method_object in self.class_methods.items():
            for call in method_object.calls:
                call.name = class_method_name
            all_class_method_calls += method_object.calls

        return all_class_method_calls

    @property
    def first_instance(self):
        if len(self.instances) == 0:
            return None

        return self.instances[0]

    @property
    def all_calls(self):

        all_calls = self.method_calls + self.attribute_calls
        all_calls = sorted(all_calls, key=lambda c: c.time)

        return all_calls

    @property
    def method_calls(self):
        all_calls = []
        all_calls += self.class_method_calls
        if self.first_instance != None: 
            all_calls += self.first_instance.method_calls

        all_calls = sorted(all_calls, key=lambda c: c.time)
        return all_calls

    @property
    def attribute_calls(self):
        all_calls = []
        for instance in self.instances:
            print instance.attribute_calls
            all_calls += instance.attribute_calls

        all_calls = sorted(all_calls, key=lambda c: c.time)

        return all_calls


def function(path,*args, **kwargs ):
    returns = kwargs.get("returns", None)
    kwargs.pop("returns", None)
    module_name, original_function_name = path.rsplit('.', 1)
    module = __import__(module_name, locals(), globals(), [original_function_name])
    #module = importlib.import_module('.'.join(path.split('.')[:-1]))
    original_function= getattr(module, original_function_name)
    function_mock = MethodMock(returns=returns)

    def wrapper(fn):
        def test_wrapped(self):
            setattr(module, original_function_name, function_mock)
            try:
                output = fn(self, function_mock)
            finally:
                pass
                setattr(module, original_function_name, original_function)
            return output
        return test_wrapped
    return wrapper

def object(path, attributes=[], methods={}):
    module_name, original_class_name = path.rsplit('.', 1)
    module = __import__(module_name, locals(), globals(), [original_class_name])
    #module = importlib.import_module('.'.join(path.split('.')[:-1]))
    original_class = getattr(module, original_class_name)
    #warnings.warn("path is " + path + " " + str(type(original_class)))

    patch = PatchObject(template=original_class, methods=methods, attributes=attributes)

    def wrapper(fn):
        def test_wrapped(self):
            setattr(module, original_class_name, patch)
            try:
                output = fn(self, patch)
            finally:
                setattr(module, original_class_name, original_class)
            return output
        return test_wrapped
    return wrapper





from distutils.core import setup

setup(name='mattys_mocks',
      version='0.1.1',
      description='Yet another mocking library',
      author='Matt ODonnell',
      author_email='matty@mavenrd.com',
      url='https://github.com/mattyo/mattysmocks',
      packages=['mattys_mocks', 'mattys_mocks.tests', 'mattys_mocks.tests.mock_these'],
     )



class DumbClass():
    a_attribute = 'test attribute'
    b_attribute = 'test b attribute'

    def a_method(self):
        pass

    def b_method(self):
        pass

    @property
    def a_property(self):
        pass

    @classmethod
    def class_method(self):
        pass
    @staticmethod
    def static_method():
        pass



from unittest import TestCase
from patch import PatchObject
from object import mock, MockObject, MethodMock, AttributeMock
from datetime import datetime
from mock_these.dumb_class import DumbClass


class TestPatchObject(TestCase):
    def setUp(self):
        self.PatchObject = PatchObject(template=DumbClass)

        an_instance = self.PatchObject()
        b_instance = self.PatchObject()

        self.instance = an_instance

        self.PatchObject.a_class_method()
        an_instance.a_method()
        an_instance.b_method()
        an_instance.a_property = "test value"
        b_instance.a_attribute = "test value"
        an_instance.a_attribute = "test value"
        b_instance.a_method()
        an_instance.a_property
        an_instance.a_attribute

    def tearDown(self):
        self.PatchObject = None

    def test_first_instance_is_none_if_none_exsist(self):
        PatchObj2 = PatchObject(template=DumbClass)
        self.assertEqual(PatchObj2.first_instance, None)


    def test_first_instance(self):
        self.assertEqual(self.PatchObject.first_instance, self.instance)

    def test_class_method_calls(self):
        self.assertEqual(len(self.PatchObject.class_method_calls), 1)

    def test_all_calls(self):
        all_call_names = map(lambda c: c.name, self.PatchObject.all_calls)
        self.assertEqual(all_call_names, 
                [   'a_class_method', 
                    'a_method',
                    'b_method',
                    'a_property',
                    'a_attribute',
                    'a_attribute',
                    'a_method',
                    'a_property',
                    'a_attribute'])

    #will produce method dups for different instances
    def test_all_method_calls(self):
        #print len(self.PatchObject.instances)
        all_method_names = map(lambda c: c.name, self.PatchObject.method_calls)
        self.assertEqual(all_method_names,
                ['a_class_method', 'a_method', 'b_method', 'a_method'])

    def test_all_attribute_calls(self):
        all_attribute_names = map(lambda c: c.name, self.PatchObject.attribute_calls)
        self.assertEqual(all_attribute_names,
                [   'a_property',
                    'a_attribute',
                    'a_attribute',
                    'a_property',
                    'a_attribute'])


class TestMockObject(TestCase):

    #these are a little harder because they involve some manual mocking
    #=================================================================================
    #def test_mock_object_should_throw_a_wargning_for_unknow_method(self):
    #    mock_object = MockObject()

    #    mock_object.somemethod()

    #    self.assertTrue(False)

    #def test_mock_object_should_throw_a_wargning_for_unknow_method_only_for_first_time_called(self):
    #    mock_object = MockObject()

    #    mock_object.somemethod()
    #    mock_object.somemethod()

    #    self.assertTrue(False)
    #=================================================================================

    #blocked on adding attributemocks in mock constructor
    #=================================================================================
    #def test_mock_object_should_accpet_methods_in_init(self):
    #    mock_object = MockObject(methods=['amethod'])
    #    self.assertIsInstance(mock_object.amethod, MethodMock)

    #def test_mock_object_should_accept_methods_in_init_in_dict_form(self):
    #    mock_object_two = MockObject(method={'amethod':'somevalue'})
    #    self.assertEqual(mock_object.amethod(), 'somevalue')

    #def test_mock_object_should_accept_attributes_in_init(self):
    #    mock_object = MockObject(attributes=['someattribute'])
    #    mock_object.someattribute = AttributeMock('someattribute', None)

    #    self.assertEqual(mock_object.someattribute, None)

    #def test_mock_object_should_accept_attributes_in_init_in_dict_form(self):
    #    mock_object = MockObject(attributes={'someattribute':'somevalue'})
    #    self.assertEqual(mock_object.someattribute, 'somevalue')
    #=================================================================================

    def test_all_calls(self):
        mock_object = mock(template=DumbClass)()
        mock_object .a_attribute = 1
        mock_object.a_method()
        get_something = mock_object.a_attribute
        mock_object.b_method()
        mock_object.b_attribute = 2


        call_names = map(lambda c: c.name, mock_object.all_calls)
        self.assertEqual(call_names,
                ['a_attribute', 'a_method', 'a_attribute', 'b_method', 'b_attribute'] )

    def test_method_calls(self):
        mock_object = mock(template=DumbClass)()
        mock_object.a_method()
        mock_object.b_method()
        self.assertEqual(len(mock_object.method_calls), 2)
        self.assertEqual(mock_object.method_calls[0].name, 'a_method')
        self.assertEqual(mock_object.method_calls[1].name, 'b_method')

    def test_attribute_calls(self):
        an_instance  = mock(tempalte=DumbClass, attributes=['a_attribute', 'b_attribute'])()

        an_instance.a_attribute = 1
        get_something = an_instance.a_attribute
        an_instance.b_attribute = 2

        attribute_names = map(lambda c: c.name, an_instance.attribute_calls)
        attribute_values = map(lambda c: (c.initial_value, c.final_value), an_instance.attribute_calls)

        self.assertEqual(attribute_names, ['a_attribute', 'a_attribute', 'b_attribute'])
        self.assertEqual(attribute_values, [(None, 1), (1,1), (None,2)])

class TestMock(TestCase):
    def setUp(self):
        self.MockClass = mock(template=DumbClass)

    def test_mock_should_return_a_mock_class(self):
        self.assertIsInstance(mock(), type(MockObject))

    def test_mock_class_should_have_default_methods_of_a_template(self):
        self.assertTrue('a_method' in dir(mock(template=DumbClass)))
        self.assertIsInstance(self.MockClass.a_method, MethodMock)


    #this test should be deleted when attributes can be safely added through the mock constructor
    def test_mock_should_add_the_attribute_attribute_names(self):
        self.assertEqual(sorted(self.MockClass.attribute_names), sorted(['a_property', 'a_attribute', 'b_attribute']))

    def test_mock_class_default_methods_should_include_class_methods(self):
        self.assertTrue('class_method' in dir(mock(template=DumbClass)))
        self.assertIsInstance(self.MockClass.class_method, MethodMock)

    def test_mock_class_default_methods_should_include_static_methods(self):
        self.assertTrue('static_method' in dir(mock(template=DumbClass)))
        self.assertIsInstance(self.MockClass.static_method, MethodMock)

    def test_mock_attribute_can_get_calls(self):
        test_mock = self.MockClass()
        test_mock.a_attribute
        self.assertEquals(len(test_mock.attribute_calls), 1)

    def test_mock_class_takes_a_list_of_methods(self):
        self.assertFalse('extra' in dir(mock(template=datetime)))
        self.assertTrue('extra' in dir(mock(template=datetime, methods=['extra'])))

    def test_mock_class_should_properties_shoud_behave_like_attributes(self):
        test_mock = self.MockClass()
        test_mock.a_property
        self.assertEqual(test_mock.a_property, None)

    def test_mock_class_undefined_methods_should_return_mock_method(self):
        test_mock = self.MockClass()
        test_mock.notamethod()
        self.assertIsInstance(test_mock.notamethod, MethodMock)

    def test_mock_class_takes_a_list_of_dicts_for_return_methods(self):
        test_mock = mock(template=datetime, methods={'extra':"test value"})()
        self.assertEqual(test_mock.extra(), "test value")

    def test_mock_class_takes_a_list_of_attributes(self):
        self.assertTrue('extra' in dir(mock(template=datetime, attributes=['extra'])))

    def test_default_return_for_method_mock_is_none(self):
        test_mock = self.MockClass()
        self.assertEqual(test_mock.a_method(), None)

    def test_mock_class_takes_a_list_of_attributes_as_dicts_for_default_values(self):
        test_mock = mock(template=datetime, attributes={'extra':"test value"})()
        self.assertEqual(test_mock.extra, "test value" )

